//
// GifController.m (c) 1998 Andrew C Stone
// andrew@stone.com (505) 345-4800  http://www.stone.com - Create(TM)
//
// THE BRAINS BEHIND GIFfun!
// Permission is granted to use this code, but please retain credit
// And as always, CAVEAT EMPTOR!
// formatted with tabs for instant web page making, make your window wide!

#import "GifController.h"

#import "GIFfun.h"
#import "GifInfo.h"
#import "GifWell.h"
#import "SDTableView.h"
#import "Preferences.subproj/GIFPrefs.h"
#import "Info.subproj/InfoAndHelp.h"
#import "TIFFtoGIF.subproj/GimmeGIF.h"
#import "SDTask.subproj/SDLogTask.h"
#import "NSData-Extensions.h"
//#import "InternetLauncher.h"

// may 6 1998 acs : added support for WIN gifAnim packages to open
#ifdef WIN32
#define OPEN_SESAME	@"openSesame.gifAnim"
#define SESAME_FILE(x) ([x stringByAppendingPathComponent:OPEN_SESAME])
#define SESAME_MESSAGE  @"<html><body>GIFfun(tm) by <A HREF=\"http://www.stone.com/\">Stone Design</A></BODY></HTML>"
#endif

//
// If you code PASCAL style (top down), then you'll
// need to prototype methods which are defined
// later in the source file after a previous usage
//

@interface GifController(ShushMyLittlePrettyCompiler)
- (void)loadDefaults;
- (void)tryAndDoEverything:(NSString *)path;
- (void)setLookAndFeelOfTableView;
- (void)readSettingsFromDict:(NSDictionary *)dict;
- (void)loadStringsIntoComboBox:(NSComboBox *)box;
- (void)setTrailingOffsetForSize:(NSSize)size window:(NSWindow *)win;
@end

@implementation NSApplication(Stone_Overrides)


- (void)orderFrontStandardAboutPanel:(id)sender {
    NSDictionary *info = [[NSBundle mainBundle] infoDictionary];
    NSString *s = [info objectForKey:@"NSAppVersion"];
    NSMutableDictionary *options = [[NSMutableDictionary alloc] init];

    if (!s || [s isEqualToString:@""]) s = [info objectForKey:@"CFBundleVersion"];
    [options setObject:s forKey:@"Version"];

    s = [info objectForKey:@"NSHumanReadableCopyright"];
    if (s) [options setObject:s forKey:@"Copyright"];

    [options setObject:@"Stone GIFfun" forKey:@"ApplicationName"];

    [[NSApplication sharedApplication] orderFrontStandardAboutPanelWithOptions:options];
    [options release];
}
@end

@implementation GifController

//
// Application Delegate methods
//
static NSString *GifInfoPasteBoard = @"GIFinfoPasteboardtype";

- (void)applicationWillFinishLaunching:(NSNotification *)notification
{
    // we are the recipient of the service calls outlined in CustomInfo.plist:
	[[notification object] setServicesProvider:self];

   // force an +initialize and thus registry...
	[GIFPrefs sharedDefaults];	

    // find the executable in our Resouces directory:
	{
		NSString *path;
#ifdef WIN32
		NSString *ext = @"exe";
#else
		NSString *ext = @"";
#endif
		if ((path = [[NSBundle mainBundle] pathForResource:@"WhirlGif" ofType:ext]))
			whirlgif = [[NSString alloc]initWithString:path];
		else NSLog(@"Cannot find WhirlGif executable...");
	}
}

//
// We make sure that we are initialized, even if launched from services or double-click
//

- (void)doInitThingsOnce
{
	static  BOOL first = YES;
	if (first) {
		first = NO;
	// let's restore the user's previous state:
		[self loadDefaults];
		[window setFrameAutosaveName:NSStringFromClass([self class])];

		[self setLookAndFeelOfTableView];
                
                [statusField setStringValue:DROP_ON_KEY];

#ifndef USING_DPS
                [drawer setPreferredEdge:NSMaxXEdge];
                [tableView registerForDraggedTypes:[NSArray arrayWithObject:GifInfoPasteBoard]];
#endif
	}
}

- (void)awakeFromNib {
   [self doInitThingsOnce];

   if (drawer) {
       toolSize = [NSWindow contentRectForFrameRect:[optionsWindow frame] styleMask:[optionsWindow styleMask]].size;
       [drawer setMinContentSize:NSMakeSize(toolSize.width - 10.0,toolSize.height - 20.0)];
       [drawer setMaxContentSize:toolSize];
        [self setTrailingOffsetForSize:[window frame].size window:window];

           [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(drawerWillBeResized:) name:NSWindowDidResizeNotification object:nil];
   }
}

- (IBAction)toggleDrawer:(id)sender {
#ifndef USING_DPS
    if ([drawer state]==NSDrawerClosedState)
        [drawer open:self];
    else if ([drawer state]==NSDrawerOpenState)
        [drawer close:self];
#endif
}
// 
// This method is all we need to implement to open files by double-clicking:
//

- (BOOL)application:(NSApplication *)sender openFile:(NSString *)filename;
{  
	// in case we don't finish launching!
	[self doInitThingsOnce];
	[self tryAndDoEverything:filename];
	return YES;
}

- (void)applicationDidFinishLaunching:(NSNotification *)notification
{
	[self doInitThingsOnce];
        [window makeKeyAndOrderFront:self];
}

//
// User Preferences
// deal with users preferences by remembering whatever they had before
// At "Create A GIF time" we save the current state
//
- (void)loadDefaults
{
	NSUserDefaults *ud = [NSUserDefaults standardUserDefaults];
        id obj = [ud objectForKey:TRANS_COLOR];
        [useColorSwitch setState:[ud boolForKey:USE_TRANS_COLOR]];
        if (obj) [transparentColorWell setColor:[NSUnarchiver unarchiveObjectWithData:obj]];

	[disposalMatrix selectCellWithTag:[ud integerForKey:DEF_DISPOSAL]];
	[loopField setIntValue:[ud integerForKey:DEF_LOOPS]];
	[loopMatrix selectCellWithTag:[ud integerForKey:DEF_LOOP_METHOD]];
        
        [minimizeOutputSwitch setState:[ud boolForKey:DEF_MINIMIZE]];
        [globalMapSwitch setState:[ud boolForKey:DEF_GLOBALMAP]];

}

//
// Factoring the code into small chunks which all methods
//  pass through the same call stack:
//

// 
// Is a file possible to be read as an image?
//

- (BOOL)isImageFile:(NSString *)path
{
    NSString *extension = [path pathExtension];
    if (!extension || [extension isEqualToString:@""]) extension = NSHFSTypeOfFile(path);
    if (extension && [[NSImage imageFileTypes] containsObject:extension]) return YES;
    return NO;
}

- (BOOL)isGifFile:(NSString *)path
{
	BOOL isDir;
	return ([[NSFileManager defaultManager] fileExistsAtPath:path isDirectory:&isDir] && !isDir && (
		([[path pathExtension] caseInsensitiveCompare:@"gif"] == NSOrderedSame)

#if !defined(OS_API)
		|| [self isImageFile:path]
#endif
	));
}


- (int)tryAndLoadPath:(NSString *)path
{
	if ([self isGifFile:path]) return [self addGifFile:path];
	else return [self loadGifsInDirectory:path];
}

- (void)tryAndDoEverything:(NSString *)path
{
	if ([self tryAndLoadPath:path]) [self mergeGifs:self];
}


// Service methods: (see CustomInfo.plist)

- (void)loadAFileOrDirectory:(NSPasteboard *)pboard userData:(NSString *)ud
                        error:(NSString **)msg
{
	NSArray *fileNames;
	int i, count;

	fileNames = [pboard propertyListForType:NSFilenamesPboardType];
	count = [fileNames count];

    // if just one FILE, we'll add it to the list:
	if (count == 1) {
		[self tryAndDoEverything:[fileNames objectAtIndex:0]];
	} else {
		for (i = 0; i < [fileNames count]; i++) {
			NSString *n = [fileNames objectAtIndex:i];
			if ([self isGifFile:n]) [self addGifFile:n];
        	}
        	[self mergeGifs:self];
	}
}


#define GIF_FILE_NAME	@"GF.gif"

- (NSString *)nextUniqueNameUsing:(NSString *)templatier
{
	static int unique = 1;
	NSString *tempName = nil;
	do {
		tempName =[NSString stringWithFormat:@"%@_%d.%@",
		[templatier stringByDeletingPathExtension],unique++,
		[templatier pathExtension]];
	} while ([[NSFileManager defaultManager] fileExistsAtPath:tempName]);

	return tempName;
}
- (NSString *)temporaryDirectory
{
	NSString *tempDir =[[NSTemporaryDirectory() 
		stringByAppendingPathComponent:
		[[NSProcessInfo processInfo] processName]]
		stringByAppendingPathComponent:NSUserName()];

    if (! [[NSFileManager defaultManager] createWritableDirectory:tempDir]) {
		NSLog(@"Couldn't create %@, using %@",tempDir, 
			NSTemporaryDirectory());
		tempDir = NSTemporaryDirectory();
	}
	return tempDir;
}

- (NSString *)scratchFolder
{
	NSString *s = [[NSUserDefaults standardUserDefaults]stringForKey:DEF_DIRECTORY];
	if (!s || [s isEqualToString:@""] ||
     ![[NSFileManager defaultManager] createWritableDirectory:s])
			s = [self temporaryDirectory];
	return s;
}

// if the file package has 00, 01, 02 endings, then we should sort accordingly:

- (NSArray *)sortFiles:(NSArray *)dp
{
	return [dp sortedArrayUsingSelector:@selector(compare:)];
}

- (int)loadGifsInDirectory:(NSString *)directory;
{
    int i;
    int count;
    NSArray *dp;
    NSString *shortname;
    NSString *path;
    NSDictionary *dict;
    NSFileManager *fileManager = [NSFileManager defaultManager];
    NSString *settingsPath;
    // start from scratch:
    [self clearGifs:self];

#ifdef WIN32
    if ([[directory lastPathComponent] isEqualToString:OPEN_SESAME])
            directory = [directory stringByDeletingLastPathComponent];
#endif
    settingsPath = [directory
                    stringByAppendingPathComponent:GIFFUN_SETTINGS_FILE];

    dp = [fileManager directoryContentsAtPath:directory];
    dp = [self sortFiles:dp];
    if ((count = [dp count])>0) {
            for (i = 0; i < count; i++) {
                    shortname = [dp objectAtIndex:i];
                    path = [directory stringByAppendingPathComponent:shortname];
                if ([self isGifFile:path]) {
                        [self addGifFile:path];
                }
        }
}
//
// Now, if we were the ones who saved these GIFS
// then we'll find a file which has our settings in dictionary form
// We'll try and load it, but we don't care if the dict is not there..
//

	if ([[NSFileManager defaultManager] isReadableFileAtPath:settingsPath] 
	 && ((dict = [NSDictionary dictionaryWithContentsOfFile:settingsPath]) != nil)) {
		[self readSettingsFromDict:dict];
	}
	return [gifFileArray count];
}

- (int)addGifFile:(NSString *)aFile
{
	NSUserDefaults *ud = [NSUserDefaults standardUserDefaults];
// On OpenStep (ie OS_API defined) this will never be the case:
	BOOL needsConversion = ([[aFile pathExtension] caseInsensitiveCompare:@"gif"] != NSOrderedSame);

	if (needsConversion) {
        // /somewhere/something.tiff -> /tmp/GIFfun/andrew/something.gif
		NSString *gifFile = [[self temporaryDirectory]
			stringByAppendingPathComponent:
			[[[aFile lastPathComponent]stringByDeletingPathExtension] 
			stringByAppendingPathExtension:@"gif"]];

		if ([[GimmeGIF sharedInstance]
			convertImageFile:aFile toGIFFile:gifFile])
			aFile = [[gifFile copy]autorelease];
		else return 0;
	}

	if (!gifFileArray)
		gifFileArray = [[NSMutableArray alloc]initWithCapacity:1];
	[gifFileArray addObject:
		[[GifInfo alloc]initWithPath:aFile
			andDelay:[ud integerForKey:DEF_DELAY]]];
	[tableView reloadData];
	[tableView scrollRowToVisible:[tableView numberOfRows]-1];

	// status field info:
	[theDirectory release];
	theDirectory = [[[aFile stringByDeletingLastPathComponent]stringByDeletingPathExtension] copy];

	[statusField setStringValue:
		[NSString stringWithFormat:@"%@: %@",DIRECTORY_KEY, theDirectory]];
	return 1;
}



#define HTML_CODE	@"<HTML><HEAD><title>%@</title></HEAD><BODY BGCOLOR=\"#000000\" TEXT=\"#FFFFFF\"><h2>%@</h2><HR><IMG SRC=\"%@\"></BODY></HTML>"

- (void)createAndOpenHTML:(NSString *)outFile
{
	NSUserDefaults *ud = [NSUserDefaults standardUserDefaults];
	NSString *htmlFile = [[outFile stringByDeletingPathExtension]stringByAppendingPathExtension:@"html"];
	NSString *s = [NSString stringWithFormat:HTML_CODE,outFile,outFile,[outFile lastPathComponent]];
        NSFileManager *fm = [NSFileManager defaultManager];
	[fm removeFileAtPath:htmlFile handler:nil]; // get rid of old
        if ([s writeToFile:htmlFile]  && [ud boolForKey:DEF_USE_BROWSER]) {
/*
#ifndef USING_DPS
    id internetConfig = [[InternetLauncher alloc] init];
    NSURL *url = [NSURL fileURLWithPath:htmlFile];
    NSLog([url path]);
    [internetConfig launchURL:[url path]];
    [internetConfig release];

#else
*/
                        [[NSWorkspace sharedWorkspace] openFile:htmlFile];
//#endif

        }
}


//
// A more complicated, interrelated set of UI items: a Text field & a radio matrix
// Therefore, we must talk to these through a higher level abstraction:
//

- (int)loopKeyValue
{
	switch ( [loopMatrix selectedTag]) {
		case LOOP_NONE: return 0;
		case LOOP_SPECIFIED: return [loopField intValue];
		case LOOP_FOREVER:
		default:
			return -1;
	}
}

- (void)setLoopFromKeyValue:(int)value
{
	switch ( value) {
		case 0: [loopMatrix selectCellWithTag:LOOP_NONE]; break;
		case -1: [loopMatrix selectCellWithTag:LOOP_FOREVER]; break;
		default:
			[loopMatrix selectCellWithTag:LOOP_SPECIFIED];
    		        [loopField setIntValue:value];
    }
}
// *************

- (NSArray *)getLoopArray
{
	switch ( [loopMatrix selectedTag]) {
		case LOOP_NONE: 
			return nil;
	
		case LOOP_SPECIFIED: 
			return [NSArray arrayWithObjects:@"-loop",
			[NSString stringWithFormat:@"%d",[loopField intValue]],nil];
		case LOOP_FOREVER:
		default:
                    return [NSArray arrayWithObject:@"-loop"];
	}
	return nil;
}

//
// IB METHODS:
//



- (NSArray *)getFileNames
{
	int count = [gifFileArray count];
	NSMutableArray *a = [NSMutableArray arrayWithCapacity:count*2];      
        int i;
       	[a addObject:[[gifFileArray objectAtIndex:0]path]];
        for (i = 1; i<[gifFileArray count]; i++) {
                GifInfo *gi = [gifFileArray objectAtIndex:i];
	               [a addObject:@"-time"];
	[a addObject:[NSString stringWithFormat:@"%d", [gi delay]]];
		[a addObject:[gi path]];
        }
        return a;
}

NSString *
colorStringFor(NSColor * color)
{
   float r,g,b;
if (color)
   [[color colorUsingColorSpaceName:NSCalibratedRGBColorSpace] getRed:&r green:&g blue:&b alpha:NULL];
else r = g = b = 1.; 	// ie WHITE
   return  [NSString stringWithFormat:@"#%02X%02X%02X", (int)(255*r), (int)(255*g), (int)(255*b)];
}


- (void)mergeGifs:(id)sender
{
        int count = [gifFileArray count];
	if (count > 0 ) {
		NSString *baseName = (theDirectory && ![theDirectory isEqualToString:@""])?
			[theDirectory lastPathComponent] : @"GIFfun";
		NSString *disp = [[disposalMatrix selectedCell]title];
		NSArray *loop = [self getLoopArray];
                NSString *colorString = colorStringFor([transparentColorWell color]);
		NSArray *fileNames = [self getFileNames];
		NSMutableArray *args = [NSMutableArray arrayWithCapacity:count*2];
		NSString *outFile;
// get our outputFile:
		[outputFile autorelease];
		outputFile = [[self nextUniqueNameUsing:[[self scratchFolder] stringByAppendingPathComponent:[baseName stringByAppendingPathExtension:@"gif"]]]retain];
                outFile = [outputFile lastPathComponent];
// get rid of old file if it exists:
               [[NSFileManager defaultManager] removeFileAtPath:outputFile handler:nil];


/// grab the options and arguments:
                if ([minimizeOutputSwitch  state]) {
                   [args addObject:@"-minimize"];
                } else if ([globalMapSwitch state]) {
                   [args addObject:@"-globmap"];
                }
                
               if ([useColorSwitch state]) {
                   [args addObject:@"-trans"];
                   [args addObject:colorString];
                }
               [args addObject:@"-time"];
[args addObject:[NSString stringWithFormat:@"%d",[[gifFileArray objectAtIndex:0]delay]]];

                [args addObject:@"-o"];
                [args addObject:outFile];
                if (loop) [args addObjectsFromArray:loop];
                [args addObject:@"-disp"];
                [args addObject:disp];
                [args addObjectsFromArray:fileNames];

// Here's my groovy Task wrapper which logs errors:

		[[SDLogTask alloc]initAndLaunchWithArgs:args executable:whirlgif
			directory:[outputFile stringByDeletingLastPathComponent] 
			 logToText:logText includeStandardOutput:NO owner:self];

	} else {
		NSBeep();
		[statusField setStringValue:DROP_ON_KEY];
	}
}

//
// Here is the callback from SDLogText
//
- (void)taskTerminated:(BOOL)success
{
	if(success) {
		[ouputDragWell setFile:outputFile];
		[statusField setStringValue:[NSString stringWithFormat:@"%@: %@",FINISHED_MSG,outputFile]];
		[self createAndOpenHTML:outputFile];
	} else [statusField setStringValue:JOB_FAILED];
}

///*********

- (void)openGifFolder:(id)sender
{
	// run open panel to allow selection of a directory
	id openpanel = [NSOpenPanel openPanel]; 

	[openpanel setTitle:@"Open Directory of GIFS"];
	[openpanel setAllowsMultipleSelection:NO];
	[openpanel setCanChooseDirectories:YES];
	if ([openpanel runModal] == NSOKButton) {  
		[self loadGifsInDirectory:[openpanel filename]];     
	}
}

- (void)addAnotherFile:(id)sender
{
	NSArray *fileTypes;
	id openpanel = [NSOpenPanel openPanel]; 

#if defined(OS_API)
	fileTypes = [NSArray arrayWithObjects:@"gif",@"GIF",nil];
	[openpanel setTitle:@"Add GIF Files..."];
#else
	fileTypes = [NSImage imageFileTypes];
	[openpanel setTitle:@"Add Image Files..."];

#endif

	[openpanel setAllowsMultipleSelection:YES];

	if ([openpanel runModalForTypes:fileTypes]) {
		int index = 0;
		NSString *directory =  (NSString *)[openpanel directory] ;
		NSArray *files = [openpanel filenames];
		int numFiles = [files count];
		while (index < numFiles) {
		/* add file to menu */
			NSString *fullName = [directory stringByAppendingPathComponent:[files objectAtIndex:index]];
			if ([self isGifFile:fullName]) [self addGifFile:fullName];
				index++;
		}
	}      
}


- (void)clearGifs:(id)sender;
{
	[gifFileArray autorelease];
	gifFileArray = [[NSMutableArray alloc]initWithCapacity:1];
	[statusField setStringValue:@""];
	[ouputDragWell setFile:nil];
	[tableView reloadData];
}


- (void)changeLoop:(id)sender;
{
	if (sender == loopField) {
		[loopMatrix selectCellWithTag:LOOP_SPECIFIED];
	} else if (sender == loopMatrix) {	// it's the matrix
		[loopField setEnabled: ([sender selectedTag] == LOOP_SPECIFIED)];
		if ([sender selectedTag] == LOOP_SPECIFIED) [loopField selectText:self];
	}
}

//
// saving
//

// these are the NSFileManger callbacks:
+ (BOOL)fileManager:(NSFileManager *)manager
       shouldProceedAfterError:(NSDictionary *)errorDict
{
	int result;
	result = NSRunAlertPanel(@"GIFfun",
		@"File operation error: %@ with file: %@",@"Proceed", @"Stop", NULL,
		[errorDict objectForKey:@"Error"], 
		[errorDict objectForKey:@"Path"]);

	if (result == NSAlertDefaultReturn)
		return YES;
	else
		return NO;
}

+ (void)fileManager:(NSFileManager *)fm willProcessPath:(NSString *)path
{
}

// 
// Save the animated GIF somewhere other than in /tmp!
//

- (void)saveAs:(id)sender;
{
	if (outputFile && ![outputFile isEqual:@""]) {
		static NSSavePanel *sp = nil;
		if (!sp) {
			sp = [[NSSavePanel savePanel]retain];
			[sp setTitle:SAVE_AS];
			[sp setDirectory:NSHomeDirectory()];
		}
		if ([sp runModalForDirectory:nil file:[outputFile lastPathComponent]]) {
			[[NSFileManager defaultManager] copyPath:outputFile
				toPath:[sp filename] handler:[self class]];
		}
	} else NSBeep();
}

//
// Save the collected GIFs and settings in a easy-to-read way
//

// define some keys which could be defined in headers for others...
#define NUM_OTHER_DICT_ITEMS	6
#define DICT_SIZE	(2* [gifFileArray count] + NUM_OTHER_DICT_ITEMS)

#define KEY_LOOP	@"Number_Of_Loops"
#define KEY_DISPOSAL	@"Disposal_Method"
#define KEY_TRANSPARENT	@"Transparent_Index"
#define KEY_TRANSPARENT_COLOR	@"Transparent_Color"

- (NSString *)pathKeyForGifInfo:(GifInfo *)gi
{
// this is too strict and not very general
// for example if we do a similar animation, but different root
// name, then are settings won't work!
// instead, we'll optimize for Create's numbering system
// and basically, write the numeric identifier of the frame
// and we'll overlook the actual path
//    return [[gi path] lastPathComponent];
	NSString *path =  [[gi path] lastPathComponent];
// get rid of "gif":
	NSString *justNum = [path stringByDeletingPathExtension];
// now grab number: ie: 00 or 01
	NSString *numIdentifier = [justNum pathExtension];
	if (!numIdentifier || [numIdentifier isEqualToString:@""]) return path;
	return numIdentifier;
}

- (NSString *)ordinalKeyForGifInfo:(GifInfo *)gi
{
	return [NSString stringWithFormat:@"%@-order",[self pathKeyForGifInfo:gi]];
}

- (void)reorderGifArrayWithDict:(NSDictionary *)dict
{
	int i,j , cnt = [gifFileArray count];
	NSMutableArray *newList = [[NSMutableArray allocWithZone:[self zone]] initWithCapacity:cnt];
	for (i = 0; i < cnt; i++) {
		for (j = 0; j < cnt; j++) {
			GifInfo *gi = [gifFileArray objectAtIndex:j];
			id number;
			if ((number = [dict objectForKey:
				[self ordinalKeyForGifInfo:gi]]) != nil) {
                		if ([number intValue] == i) {
					[newList addObject:gi];
					break;
				}
			}
		}
	}

	if ([newList count] == cnt) {
		[gifFileArray autorelease];
		gifFileArray = newList;
		[tableView reloadData];
	} else [newList autorelease];
}


- (void)writeSettingsToDict:(NSMutableDictionary *)dict
{
	int i;
// write the delays between frames:
// and the position of the GIF in the animation
// Note that we only write out strings, arrays, data or dicts in order that
// the data can be serialized...

	for (i = 0; i < [gifFileArray count]; i++) {
		GifInfo *gi = [gifFileArray objectAtIndex:i];
		[dict setObject:[NSString stringWithFormat:@"%d",
			[gi delay]] forKey:[self pathKeyForGifInfo:gi]];
		[dict setObject:[NSString stringWithFormat:@"%d", i]
			forKey:[self ordinalKeyForGifInfo:gi]];
	}
	[dict setObject:[NSString stringWithFormat:@"%d",[self loopKeyValue]]
			forKey:KEY_LOOP];
	[dict setObject:[NSString stringWithFormat:@"%d",
			[disposalMatrix selectedTag]] forKey:KEY_DISPOSAL];
        if ([useColorSwitch state])
            [dict setObject:[NSArchiver archivedDataWithRootObject:[transparentColorWell color]] forKey:KEY_TRANSPARENT_COLOR];
}

- (void)readSettingsFromDict:(NSDictionary *)dict
{
	int i;
	BOOL needToReorder = NO;
// Are the GIFS in the correct order?
// Each GIFs ordinality key should match its position...
// but we are lazy, and we'll do this only if necessary

// read the delays between frames:
	for (i = 0; i < [gifFileArray count]; i++) {
		GifInfo *gi = [gifFileArray objectAtIndex:i];
		id number;
		if ((number = [dict objectForKey:
			[self pathKeyForGifInfo:gi]]) != nil)
			[gi setDelay:[number intValue]];
		if ((number = [dict objectForKey:
			[self ordinalKeyForGifInfo:gi]]) != nil)
			if ([number intValue] != i) needToReorder = YES;
	}


	[self setLoopFromKeyValue:[[dict objectForKey:KEY_LOOP]intValue]];
	[disposalMatrix selectCellWithTag:[[dict objectForKey:KEY_DISPOSAL]intValue]];

        {
            id obj = [dict objectForKey:KEY_TRANSPARENT_COLOR];
            if (obj) {
                [useColorSwitch setState:YES];
                [transparentColorWell setColor:[NSUnarchiver unarchiveObjectWithData:obj]];
            } else {
		[useColorSwitch setState:NO];
            }
        }
// OK - do the reordering work if you have to...
	if (needToReorder) [self reorderGifArrayWithDict:dict];

}

//
// save settings
//

- (BOOL)saveDictToFile:(NSString *)file
{
        NSMutableDictionary *dict = [NSMutableDictionary dictionaryWithCapacity:DICT_SIZE];
// write a dictionary containing our settings to this directory
        [self writeSettingsToDict:dict];
#ifdef WIN32
/* put our double-click file in there */
        [SESAME_MESSAGE writeToFile:SESAME_FILE([file stringByDeletingLastPathComponent])];

#endif
        return [dict writeToFile:file];
}

- (void)saveSettings:(id)sender;
{
	if ([gifFileArray count]) {
		static NSSavePanel *sp = nil;
		NSString *dir = [[[gifFileArray objectAtIndex:0] path]
			stringByDeletingLastPathComponent];
		if (!sp) {
			// retain it, or you'll crash on the second invoke!
			sp = [[NSSavePanel savePanel]retain];
			[sp setTitle:SAVE_GIF_SETTINGS];
			[sp setDirectory:NSHomeDirectory()];
			[sp setRequiredFileType:GIFFUN_SETTINGS_FILETYPE];
		}
        // we'll use the last directory selected
       // but we'll use the basis of the file name to be
       // the original name that was loaded:
		if ([sp runModalForDirectory:dir file:GIFFUN_SETTINGS_FILE]) {
			[self saveDictToFile:[sp filename]];
		}
	}
}

- (BOOL)writeGIFSToPath:(NSString *)destPath
{
	// Create the destination directory
	if (directoryOK(destPath)) {
		int i;
                
                {
            NSString *contents = [destPath stringByAppendingPathComponent:@"Contents"];
            // write our package info:
#warning PKGINFO haque
            if (directoryOK(contents))
            [@"gAnmgFun" writeToFile:[contents stringByAppendingPathComponent:@"PkgInfo"]];
            }
	// for each gif info:
	// copy to the destination directory
	// but only if it doesn't exist already!
		for (i = 0; i < [gifFileArray count]; i++) {
			GifInfo *gi = [gifFileArray objectAtIndex:i];
			NSString *gifName = [[gi path] lastPathComponent];
			NSFileManager *fm = [NSFileManager defaultManager];
			NSString *destFile = [destPath stringByAppendingPathComponent:gifName];
			if (![fm fileExistsAtPath:destFile])
				[fm copyPath:[gi path]
					toPath:destFile
					handler:[self class]];	
		}
		return [self saveDictToFile:
		 [destPath stringByAppendingPathComponent:GIFFUN_SETTINGS_FILE]];
	}
	return NO;
}

- (void)saveGifAnimDirectory:(id)sender;
{
	if ([gifFileArray count]) {
		static NSSavePanel *sp = nil;
		NSString *filename = [[[[gifFileArray objectAtIndex:0] path]
			stringByDeletingLastPathComponent] lastPathComponent];
		if (!sp) {
			sp = [[NSSavePanel savePanel]retain];
			[sp setTitle:SAVE_GIF_ANIMS];
			[sp setDirectory:NSHomeDirectory()];
			[sp setRequiredFileType:GIFFUN_FILE_TYPE];
		}
         // we'll use the last directory selected
	// but we'll use the basis of the file name to be
	// the original name that was loaded:
		if ([sp runModalForDirectory:nil file:filename]) {
			[self writeGIFSToPath:[sp filename]];
		}
	}
}

//
// Covers for other nib files actions:
//

- (void)runDefaultsPanel:(id)sender;
{
	[[GIFPrefs sharedDefaults] orderFront:self];
}

- (void)loadHelp:(id)sender
{
	[[InfoAndHelp sharedHelp] orderFrontHelp:self];
}

- (void)loadWhirlGifHelp:(id)sender;
{
	[[InfoAndHelp sharedHelp] orderFrontWhirlHelp:self];
}

- (void)loadSample:(id)sender;
{
	NSString *path = [[NSBundle mainBundle] pathForResource:SAMPLE_GIFS ofType:GIFFUN_FILE_TYPE];
	[self loadGifsInDirectory:path];
        [self mergeGifs:self];

}

//
// Setting up the TableView the way we want it:
//

#define ROW_HEIGHT  48

- (id)comboCell {
    NSComboBoxCell *comboCell;
    comboCell = [[NSComboBoxCell allocWithZone:[self zone]] init];
    [comboCell addItemWithObjectValue:@"10"];
    [comboCell addItemWithObjectValue:@"20"];
    [comboCell addItemWithObjectValue:@"50"];
    [comboCell addItemWithObjectValue:@"100"];
    [comboCell addItemWithObjectValue:@"200"];
    [comboCell addItemWithObjectValue:@"500"];
    [comboCell setNumberOfVisibleItems:6];
    [comboCell setEditable:YES];
    [comboCell setEnabled:YES];
    [comboCell setBezeled:YES];
    [comboCell setFont:[NSFont userFontOfSize:14.0]];
//    [[NSNotificationCenter defaultCenter] addObserver:self
//        selector:@selector(comboBoxChangedSelection:)
//        name:NSComboBoxSelectionDidChangeNotification
//        object:combo];

    return comboCell;
}

- (void)setLookAndFeelOfTableView
{
	NSImageCell *ic = [[NSImageCell allocWithZone:[self zone]] initImageCell:nil];
	NSTableColumn *tc = [tableView tableColumnWithIdentifier:I_IMAGE];
//        NSComboBox *combo;
//	NSRect r = NSZeroRect;

// Make the tableview constrain movement verticallly:
#ifdef USING_DPS
        [tableView setConstrainedMove:YES];
#endif
// Make the rows taller so we can see our GIFS:
        [tableView setRowHeight:ROW_HEIGHT];

   // We want the GIF not to be stretched, and we set the TableColumn's cell to our new ImageCell:
	[ic setImageScaling:NSScaleProportionally]; // or NSScaleToFit
	[tc setDataCell:ic];

   // the Delay field is the only editable one
   // We'll show that with a bigger font and a bezeled field:

	tc = [tableView tableColumnWithIdentifier:I_DELAY];
        [tc setDataCell:[self comboCell]];

}


//
// TableView delegate methods:
//

- (int)numberOfRowsInTableView:(NSTableView *)tv
{
	return [gifFileArray count];
}

- (NSString *)uiPath:(GifInfo *)gi
{
	return [[[gi path]lastPathComponent]stringByDeletingPathExtension];
}

- tableView:(NSTableView *)tv
        objectValueForTableColumn:(NSTableColumn *)tc
        row:(int)row
{
	NSString *ident = [tc identifier];
	if (tv == tableView) {
	if ([ident isEqual:I_NUMBER]) return [NSNumber numberWithInt:row + 1];
	else if ([ident isEqual:I_IMAGE])
		return [[gifFileArray objectAtIndex:row]image];
	else if ([ident isEqual:I_PATH])
		return [self  uiPath:[gifFileArray objectAtIndex:row]];
	else if ([ident isEqual:I_DELAY])
		return [NSNumber numberWithInt:[[gifFileArray objectAtIndex:row]delay]];
	}
	return @"";
}

- (void)tableView:(NSTableView *)tv setObjectValue:(id)object forTableColumn:(NSTableColumn *)tc row:(int)row;
{
	NSString *ident = [tc identifier];
	if (tv == tableView) {
		if ([ident isEqual:I_DELAY])
			[[gifFileArray objectAtIndex:row]setDelay:[object intValue]];
	}
}

// COMBO BOX STUFF
// these next two methods should be changed together!
- (int)valueForIndex:(int)index
{
    switch (index) {
	case 0: return 0;
	case 1: return 10;
	case 2: return 25;
	case 3: return 50;
	case 4: return 75;
	case 5: return 100;
	case 6: return 150;
	case 7: return 200;
	case 8: return 300;
	case 9: return 500;
	default: return 20;
    }
    return 20;
}

- (void)loadStringsIntoComboBox:(NSComboBox *)box
{
    [box addItemWithObjectValue:@"Fast as possible"];
    [box addItemWithObjectValue:@"Tenth of a second"];
    [box addItemWithObjectValue:@"Quarter of a second"];
    [box addItemWithObjectValue:@"Half a second"];
    [box addItemWithObjectValue:@"Three quarters second"];
    [box addItemWithObjectValue:@"Second"];
    [box addItemWithObjectValue:@"Second and a half"];
    [box addItemWithObjectValue:@"Two seconds"];
    [box addItemWithObjectValue:@"Three seconds"];
    [box addItemWithObjectValue:@"Five seconds"];
}


- (void)comboBoxChangedSelection:(NSNotification *)notification
{
    NSComboBox *combo = [notification object];
    int which = [combo indexOfSelectedItem];
    [combo performSelector:@selector(setIntValue:) withObject:[NSNumber numberWithInt:[self valueForIndex:which]] afterDelay:.1];
}

//
// SDMovingRowsProtocol for reordering of rows:
//
- (BOOL)rowOK:(int)row
{
	return (row > -1 && row < [gifFileArray count]);
}

- (unsigned int)dragReorderingMask:(int)col;
{
	// DRAG_ALWAYS means do it with no modifier needed at all
	if (col == IMAGE_COLUMN_NUMBER)  	return DRAG_ALWAYS;
	else return NSCommandKeyMask; 
}
/*
- (void)willMoveRow:(int)row;
{
	[statusField setStringValue:@""];
}

- (NSImage *)imageForRow:(int)row;
{
	if ([self rowOK:row]) {
		GifInfo *gi = [gifFileArray objectAtIndex:row]; 
		[statusField setStringValue:
			[NSString stringWithFormat:@"%@: %@", NOW_MOVING_MSG,
			 [self uiPath:gi]]];
		return [gi image];
	}
	return nil;
}
*/

-  (BOOL)tableView:(NSTableView *)tv didDepositRow:(int)rowToMove at:(int)newPosition
{
	if (rowToMove != -1 && newPosition != -1) {
		id object = [gifFileArray objectAtIndex:rowToMove];
		if (newPosition < [gifFileArray count] - 1) {
			[gifFileArray removeObjectAtIndex:rowToMove];
			[gifFileArray insertObject:object atIndex:newPosition];
		} else {
			[gifFileArray removeObjectAtIndex:rowToMove];
			[gifFileArray addObject:object];
		}
		[statusField setStringValue:[NSString stringWithFormat:
			@"%@ %@ %@ %d %@ %d", MOVED_MSG,[self uiPath:object],
			FROM_MSG,rowToMove + 1,DEST_MSG,newPosition + 1]];   

		[window setDocumentEdited:YES];
		return YES;	// ie reload
	}
	return NO;
}

// give the delegate a chance to disallow a row move
- (BOOL)tableView:(SDTableView *)tableView willMoveRow:(int)row;
{
/*
   if ([self rowOK:row]) {
       GifInfo *gi = [gifFileArray objectAtIndex:row];
       [statusField setStringValue:[NSString stringWithFormat:@"%@: %@", NOW_MOVING_MSG,
               [self uiPath:gi]]];
  }
*/
    [statusField setStringValue:@""];
   return YES;
}

// give the delegate a chance to provide an image for the moving row
- (NSImage *)tableView:(SDTableView *)aTableView imageForRow:(int)row;
{
   /*
   if ([self rowOK:row]) {
       GifInfo *gi = [gifFileArray objectAtIndex:row];
       [statusField setStringValue:[NSString stringWithFormat:@"%@: %@", NOW_MOVING_MSG,
               [self uiPath:gi]]];
        return [gi image];
  }
    */
    [statusField setStringValue:@""];
  return nil;
}

- (BOOL) tableView:(SDTableView *) tableView draggingRow:(int)draggedRow overRow:(int) targetRow
{
   return YES;
}

- (BOOL) tableView:(SDTableView *)tableView willDepositRow:(int) draggedRow at:(int) destinationRow
{
   return YES;
}

// NEW API for Dragging in TableView:
// typedef enum { NSTableViewDropOn, NSTableViewDropAbove } NSTableViewDropOperation;
// In drag and drop, used to specify a dropOperation. For example, given a table with N rows (numbered with row 0 at the top visually), a row of N-1 and operation of NSTableViewDropOn would specify a drop on the last row. To specify a drop below the last row, one would use a row of N and NSTableViewDropAbove for the operation.

static int _moveRow = 0;

- (BOOL)tableView:(NSTableView *)tv writeRows:(NSArray*)rows toPasteboard:(NSPasteboard*)pboard 
// This method is called after it has been determined that a drag should begin, but before the drag has been started. To refuse the drag, return NO. To start a drag, return YES and place the drag data onto the pasteboard (data, owner, etc...). The drag image and other drag related information will be set up and provided by the table view once this call returns with YES. The rows array is the list of row numbers that will be participating in the drag.
{
    int count = [gifFileArray count];
    int rowCount = [rows count];
    if (count < 2) return NO;  // except they might want to copy to another window?
    // we should allow group selection and copy between windows: PENDING
    [pboard declareTypes:[NSArray arrayWithObject:GifInfoPasteBoard] owner:self];
    [pboard setPropertyList:rows forType:GifInfoPasteBoard];
    if (rowCount == 1) _moveRow = [[rows objectAtIndex:0]intValue];
    return YES;
}
- (unsigned int)tableView:(NSTableView*)tv validateDrop:(id <NSDraggingInfo>)info proposedRow:(int)row proposedDropOperation:(NSTableViewDropOperation)op 
// This method is used by NSTableView to determine a valid drop target. Based on the mouse position, the table view will suggest a proposed drop location. This method must return a value that indicates which dragging operation the data source will perform. The data source may "re-target" a drop if desired by calling setDropRow:dropOperation: and returning something other than NSDragOperationNone. One may choose to re-target for various reasons (eg. for better visual feedback when inserting into a sorted position).
{
    if (row != _moveRow) {
        if (op==NSTableViewDropAbove /*NSTableViewDropOn*/) {
            return NSDragOperationAll;
        }
        return NSDragOperationNone;
    }
    return NSDragOperationNone;
}

- (BOOL)tableView:(NSTableView*)tv acceptDrop:(id <NSDraggingInfo>)info row:(int)row dropOperation:(NSTableViewDropOperation)op 
// This method is called when the mouse is released over an outline view that previously decided to allow a drop via the validateDrop method. The data source should incorporate the data from the dragging pasteboard at this time. 
{
   BOOL result = [self tableView:tableView didDepositRow:_moveRow at:(int)row];
   [tableView reloadData];
   return result;
}

// end TableView dragging API

// cut copy paste
- (void)copy:sender
{
	int toCopy = [tableView selectedRow];
	if ([self rowOK:toCopy])
		copiedRow = [[gifFileArray objectAtIndex:toCopy]copy];
	else {
		NSBeep();
		[statusField setStringValue:
			[gifFileArray count]>0?SELECT_ROW_KEY:DROP_ON_KEY];
	}
}

- (void)cut:sender
{
	int toCut = [tableView selectedRow];
	if ([self rowOK:toCut]) {
		copiedRow = [[gifFileArray objectAtIndex:toCut]copy];
		[gifFileArray removeObjectAtIndex:toCut];
		[tableView reloadData];
	}else {
		NSBeep();
		[statusField setStringValue:[gifFileArray count]>0?SELECT_ROW_KEY:DROP_ON_KEY];
    }

}

- (void)paste:sender
{
	int location = [tableView selectedRow];
	int count = [gifFileArray count];
	if (copiedRow != nil) {
		if (location == -1 || location >= count - 1)
			[gifFileArray addObject:[copiedRow copy]];
		else
			[gifFileArray insertObject:[copiedRow copy] atIndex:location+1];
		[tableView reloadData];
	} else {
		NSBeep();
		[statusField setStringValue:
			[gifFileArray count]>0?SELECT_ROW_KEY:DROP_ON_KEY];
	}
}
/*

#define REFRESH_STRING @"<HTML>\n<HEAD>\n<META HTTP-EQUIV=\"refresh\" CONTENT=\"0; URL=%@\">\n<TITLE>Switching to %@...</TITLE>\n</HEAD>\n<BODY>\n<H1>Let's turn to Stone!</H1>\nThis page should soon be replaced by %@. If this does not work or if you are of the impatient type, please click to go to <A HREF=\"%@\">%@</A>.\n</BODY>\n</HTML>"

- (NSString *)refreshPageForAddress:(NSString *)address {

    return [NSString stringWithFormat:REFRESH_STRING,address,address,address,address,address];
}
- (IBAction)goToStoneWebSite:(id)sender {
 // make a file
    // open with workspace
    // let browser do work:
    NSString *address = [self refreshPageForAddress:@"http://www.stone.com"];
    NSString *file = [self nextUniqueNameUsing:[NSTemporaryDirectory() stringByAppendingPathComponent:@"Stone.html"]];
    [address writeToFile:file atomically:YES];
    [[NSWorkspace sharedWorkspace] openFile:file];

}
*/

- (void)launchURL:(NSString *)url {
   [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:url]];
   //    [[[[InternetLauncher alloc] init] autorelease] launchURL:url];
}

- (IBAction)goToStoneWebSite:(id)sender {
    [self launchURL:@"http://www.stone.com"];
}

- (IBAction)showSuggestionWindow:(id)sender {
    [self launchURL:@"http://www.stone.com/Feedback.html"];
//    [[self sharedInstanceOfClassName:@"Suggestion"] showWindow:self];
}


// So we ask the document's windowController to call us from their willResize:toSize:
#define TOOL_WINDOW_HGT_FUDGE	40.0

- (NSSize)parentWindowWillResize:(NSWindow *)parentWindow toSize:(NSSize)size{
   if (parentWindow == [drawer parentWindow]) {
       float minimum = toolSize.height + TOOL_WINDOW_HGT_FUDGE;
       if (size.height < minimum) size.height = minimum;
   }
   return size;
}

#define MIN_WIDTH 450.0

- (NSSize)windowWillResize:(NSWindow *)sender toSize:(NSSize)size
{
    if (size.width < MIN_WIDTH) size.width = MIN_WIDTH;
    
    size = [self parentWindowWillResize:window toSize:size];
    return size;
}



- (void)setTrailingOffsetForSize:(NSSize)size window:(NSWindow *)win {
    //NSRect wr = [NSWindow contentRectForFrameRect:NSMakeRect(0.0,0.0,size.width,size.height) styleMask:[window styleMask]];
    float desired = toolSize.height + TOOL_WINDOW_HGT_FUDGE;
    float dy = MAX(0.0, [win frame].size.height - desired );
    [drawer setTrailingOffset:dy];
}

- (void)drawerWillBeResized:(NSNotification *)notification {
    NSWindow *win = [notification object];
    if (win == [drawer parentWindow]) {
        // window was resized - now we will be resized!
        [self setTrailingOffsetForSize:[win frame].size window:win];
    }
}

@end
