
#import <Foundation/Foundation.h>
#include <sys/mount.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/attr.h>
#include <fcntl.h>

@implementation NSFileManager(Stone_Extensions)

//
// Temporary Directory stuff: useful code.
//

BOOL directoryOK(NSString *path)
{
       BOOL isDirectory;
       NSFileManager *fileManager = [NSFileManager defaultManager];
       if (![fileManager fileExistsAtPath:path isDirectory:&isDirectory] || !isDirectory) {
               NSDictionary *dict = [NSDictionary dictionaryWithObject:
                       [NSNumber numberWithUnsignedLong:0777]
                       forKey:NSFilePosixPermissions];
               if (![fileManager createDirectoryAtPath:path attributes:dict]) 
                       return NO;
       }
       return YES;
}

NSString * existingPath(NSString *path)
{
       while (path && ![path isEqualToString:@""]
               && ![[NSFileManager defaultManager] fileExistsAtPath:path])
               path = [path stringByDeletingLastPathComponent];
       return path;	
}

NSArray *directoriesToAdd(NSString *path, NSString *existing)
{
       NSMutableArray *a = [NSMutableArray arrayWithCapacity:4];
       if (path != nil && existing != nil) {
               while (![path isEqualToString:existing]) {
                       [a insertObject:[path lastPathComponent] atIndex:0];
                       path = [path stringByDeletingLastPathComponent];
               }
       }
       return a;
}

// this will go up the path until it finds an existing directory
// and will add each subpath and return YES if succeeds, NO if fails:

- (BOOL)createWritableDirectory:(NSString *)path
{
       BOOL isDirectory;
       NSFileManager *fileManager = [NSFileManager defaultManager];
       if ([fileManager fileExistsAtPath:path isDirectory:&isDirectory]
               && isDirectory && [fileManager isWritableFileAtPath:path])
               return YES; // no work to do
       else {
               NSString *existing = existingPath(path);
               NSArray *dirsToAdd = directoriesToAdd(path,existing);
               int i;
               BOOL good = YES;
               for (i = 0; i < [dirsToAdd count]; i++) {
                       existing = [existing stringByAppendingPathComponent:
                               [dirsToAdd objectAtIndex:i]];
                       if (!directoryOK(existing)) {
                               good = NO;
                               break;
                       }
               }
               return good;
       }
}

- (NSString *)temporaryDirectory
{
   NSString *tempDir =[[NSTemporaryDirectory() stringByAppendingPathComponent:[[NSProcessInfo processInfo] processName]]stringByAppendingPathComponent:NSUserName()];

   if (! [self createWritableDirectory:tempDir]) {
       NSLog(@"Couldn't create %@, using %@",tempDir, NSTemporaryDirectory());
       tempDir = NSTemporaryDirectory();
   }
  return tempDir;
}

 typedef struct {
     long type;
     long creator;
     short flags;
     short locationV;
     short locationH;
     short fldr;
     short iconID;
     short unused[3];
     char script;
     char xFlags;
     short comment;
     long putAway;
} OFFinderInfo;

- (int)getType:(unsigned long *)typeCode andCreator:(unsigned long *)creatorCode forPath:(NSString *)path;
{
#ifndef USING_DPS
    struct attrlist attributeList;
    struct {
        long ssize;
        OFFinderInfo finderInfo;
    } attributeBuffer;
    int errorCode;

    attributeList.bitmapcount = ATTR_BIT_MAP_COUNT;
    attributeList.reserved = 0;
    attributeList.commonattr = ATTR_CMN_FNDRINFO;
    attributeList.volattr = attributeList.dirattr = attributeList.fileattr = attributeList.forkattr = 0;
    memset(&attributeBuffer, 0, sizeof(attributeBuffer));

    errorCode = getattrlist([self fileSystemRepresentationWithPath:path], &attributeList, &attributeBuffer, sizeof(attributeBuffer), 0);
    if (errorCode == -1) {
        switch (errno) {
            case EOPNOTSUPP:
                *typeCode = 0; // We could use the Mac APIs, or just read the "._" file.
                *creatorCode = 0;
            default:
                return (errorCode);
        }
    } else {
        *typeCode = attributeBuffer.finderInfo.type;
        *creatorCode = attributeBuffer.finderInfo.creator;
    }

    return (errorCode);
#else
     return -1;
#endif
}

- (int)setType:(unsigned long)typeCode andCreator:(unsigned long)creatorCode forPath:(NSString *)path;
{
#ifndef USING_DPS
     struct attrlist attributeList;
     struct {
         long ssize;
         OFFinderInfo finderInfo;
     } attributeBuffer;
     int errorCode;

     attributeList.bitmapcount = ATTR_BIT_MAP_COUNT;
     attributeList.reserved = 0;
     attributeList.commonattr = ATTR_CMN_FNDRINFO;
     attributeList.volattr = attributeList.dirattr = attributeList.fileattr = attributeList.forkattr = 0;
     memset(&attributeBuffer, 0, sizeof(attributeBuffer));

     getattrlist([self fileSystemRepresentationWithPath:path], &attributeList, &attributeBuffer, sizeof(attributeBuffer), 0);

     attributeBuffer.finderInfo.type = typeCode;
     attributeBuffer.finderInfo.creator = creatorCode;

     errorCode = setattrlist([self fileSystemRepresentationWithPath:path], &attributeList, &attributeBuffer.finderInfo, sizeof(OFFinderInfo), 0);
     if (errorCode == 0)
         return 0;

     if (errno == EOPNOTSUPP) {
#define MAGIC_HFS_FILE_LENGTH 82
         unsigned char magicHFSFileContents[MAGIC_HFS_FILE_LENGTH] = {
             0x00, 0x05, 0x16, 0x07, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
             0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x09, 0x00, 0x00,
             0x00, 0x32, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x52, 0x00, 0x00,
             0x00, 0x00, 't', 'y', 'p', 'e', 'c', 'r', 'e', 'a', 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
             0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
             0x00, 0x00};
         unsigned int offsetWhereOSTypesAreStored = 50;
         NSData *data;
         NSString *magicHFSFilePath;

         *((int *)(&magicHFSFileContents[offsetWhereOSTypesAreStored])) = typeCode;
         *((int *)(&magicHFSFileContents[offsetWhereOSTypesAreStored + 4])) = creatorCode;
         data = [NSData dataWithBytes:magicHFSFileContents length:MAGIC_HFS_FILE_LENGTH];
         magicHFSFilePath = [[path stringByDeletingLastPathComponent] stringByAppendingPathComponent:[@"._" stringByAppendingString:[path lastPathComponent]]];

         if ([self createFileAtPath:magicHFSFilePath contents:data attributes:[self fileAttributesAtPath:path traverseLink:NO]])
             return 0;
         else
             return errorCode;
     }
     return errorCode;
#else
     return -1;
#endif
}

- (unsigned long)codeFromString:(NSString *)fourLetterWord {
     unsigned const char *c = [fourLetterWord cString];
    if ([fourLetterWord cStringLength] == 4) {
        return ((c[0]<<24) + (c[1]<<16) + (c[2]<<8) + c[3]);
    }
    return 0;
}

- (NSString *)stringFromCode:(unsigned long)code {
    unsigned char c[4];
    c[0] = code >> 24;
    c[1] = (code >> 16) & 0xff;
    c[2] = (code >> 8) & 0xff;
    c[3] = code & 0xff;
    return  [NSString stringWithCString:c];
}

- (int)setTypeString:(NSString *) type andCreatorString:(NSString *)creator forPath:(NSString *)path {
    return [self setType:[self codeFromString:type] andCreator:[self codeFromString:creator]  forPath:path];
}

- (int)getTypeString:(NSString **)type andCreatorString:(NSString **)creator forPath:(NSString *)path {
    unsigned long t,c;
    int result = [self getType:&t andCreator:&c forPath:path];
    if (result == 0) {
        *type = [self stringFromCode:t];
        *creator = [self stringFromCode:c];
    } else {
        *type = @"";
        *creator = @"";
    }
   // NSLog(@"type = '%@'\tcreator = '%@'",*type,*creator);
    return result;
}

@end